_tinkerforge_device()
{
	local prev

	for (( i=$1; i > 0; i-- )); do
		prev="${COMP_WORDS[i-1]}"

		case "${prev}" in
		get-*|is-*|are-*|acceleration|air-pressure|air-pressure-reached|all-data|altitude|altitude-reached|ambient-temperature|ambient-temperature-reached|analog-value|analog-value-reached|angular-velocity|beep-finished|button-pressed|button-released|coordinates|current|current-reached|current-velocity|date-time|distance|distance-reached|emergency-shutdown|humidity|humidity-reached|illuminance|illuminance-reached|interrupt|magnetic-field|monoflop-done|morse-code-finished|motion|new-state|object-temperature|object-temperature-reached|orientation|over-current|position|position-reached|power|power-reached|pressed|quaternion|released|resistance|resistance-reached|stack-current|stack-current-reached|stack-voltage|stack-voltage-reached|status|temperature|temperature-reached|under-voltage|usb-voltage|usb-voltage-reached|velocity-reached|voltage|voltage-reached)
			COMPREPLY=($(compgen -W "--help --execute" -- ${cur}))
			break
			;;
		set-*|backlight-off|backlight-on|beep|calibrate|clear-display|disable|disable-position-reached-callback|disable-velocity-reached-callback|drive-backward|drive-forward|enable|enable-position-reached-callback|enable-velocity-reached-callback|full-brake|leds-off|leds-on|morse-code|orientation-calculation-off|orientation-calculation-on|refresh-wifi-status|reset|restart|stop|write-line)
			COMPREPLY=($(compgen -W "--help --expect-response" -- ${cur}))
			break
			;;
		call)
			local functions=$(tinkerforge call $2 --list-functions)
			COMPREPLY=($(compgen -W "--help --list-functions ${functions}" -- ${cur}))
			break
			;;
		dispatch)
			local callbacks=$(tinkerforge dispatch $2 --list-callbacks)
			COMPREPLY=($(compgen -W "--help --list-callbacks ${callbacks}" -- ${cur}))
			break
			;;
		*)
			;;
		esac
	done
}

_tinkerforge()
{
	COMPREPLY=()
	local prev cur
	prev="${COMP_WORDS[COMP_CWORD-1]}"
	cur="${COMP_WORDS[COMP_CWORD]}"

	case "${prev}" in
	--help)
		return 0
		;;
	--version)
		return 0
		;;
	--host)
		_known_hosts
		return 0
		;;
	--port)
		return 0
		;;
	--item-separator)
		return 0
		;;
	--group-separator)
		return 0
		;;
	--no-symbolic-input)
		return 0
		;;
	--no-symbolic-output)
		return 0
		;;
	--duration)
		return 0
		;;
	--timeout)
		return 0
		;;
	--list-devices)
		return 0
		;;
	--list-functions)
		return 0
		;;
	--list-callbacks)
		return 0
		;;
	ambient-light-bricklet|analog-in-bricklet|analog-out-bricklet|barometer-bricklet|current12-bricklet|current25-bricklet|dc-brick|distance-ir-bricklet|dual-relay-bricklet|gps-bricklet|humidity-bricklet|imu-brick|industrial-digital-in-4-bricklet|industrial-digital-out-4-bricklet|industrial-dual-0-20ma-bricklet|industrial-quad-relay-bricklet|io16-bricklet|io4-bricklet|joystick-bricklet|lcd-16x2-bricklet|lcd-20x4-bricklet|linear-poti-bricklet|master-brick|piezo-buzzer-bricklet|ptc-bricklet|rotary-poti-bricklet|servo-brick|stepper-brick|temperature-bricklet|temperature-ir-bricklet|voltage-bricklet|voltage-current-bricklet)
		local host port options
		host=""
		port=""
		options=""

		for (( i=${COMP_CWORD}; i > 0; i-- )); do
			case "${COMP_WORDS[i]}" in
			--host)
				case "${COMP_WORDS[i+1]}" in
				=)
					host="--host ${COMP_WORDS[i+2]}"
					;;
				*)
					host="--host ${COMP_WORDS[i+1]}"
					;;
				esac
				;;
			--port)
				case "${COMP_WORDS[i+1]}" in
				=)
					port="--port ${COMP_WORDS[i+2]}"
					;;
				*)
					port="--port ${COMP_WORDS[i+1]}"
					;;
				esac
				;;
			call)
				options="--list-functions"
				;;
			dispatch)
				options="--list-callbacks"
				;;
			*)
				;;
			esac
		done

		local uids=$(tinkerforge ${host} ${port} enumerate --duration 100 --execute "if [ {device-identifier} = '${COMP_WORDS[COMP_CWORD-1]}' ]; then echo {uid}; fi" 2> /dev/null)
		COMPREPLY=($(compgen -W "--help ${options} ${uids}" -- ${cur}))
		return 0
		;;
	--execute)
		return 0
		;;
	--expect-response)
		return 0
		;;
	*)
		;;
	esac

	for (( i=${COMP_CWORD}; i > 0; i-- )); do
		prev="${COMP_WORDS[i-1]}"

		case "${prev}" in
		tinkerforge)
			COMPREPLY=($(compgen -W "--help --host --port --item-separator --group-separator --no-symbolic-input --no-symbolic-output call dispatch enumerate" -- ${cur}))
			break
			;;
		call)
			local devices=$(tinkerforge call --list-devices)
			COMPREPLY=($(compgen -W "--help --list-devices --timeout ${devices}" -- ${cur}))
			break
			;;
		dispatch)
			local devices=$(tinkerforge dispatch --list-devices)
			COMPREPLY=($(compgen -W "--help --list-devices --duration ${devices}" -- ${cur}))
			break
			;;
		enumerate)
			COMPREPLY=($(compgen -W "--help --duration --types --execute" -- ${cur}))
			break
			;;
		ambient-light-bricklet|analog-in-bricklet|analog-out-bricklet|barometer-bricklet|current12-bricklet|current25-bricklet|dc-brick|distance-ir-bricklet|dual-relay-bricklet|gps-bricklet|humidity-bricklet|imu-brick|industrial-digital-in-4-bricklet|industrial-digital-out-4-bricklet|industrial-dual-0-20ma-bricklet|industrial-quad-relay-bricklet|io16-bricklet|io4-bricklet|joystick-bricklet|lcd-16x2-bricklet|lcd-20x4-bricklet|linear-poti-bricklet|master-brick|piezo-buzzer-bricklet|ptc-bricklet|rotary-poti-bricklet|servo-brick|stepper-brick|temperature-bricklet|temperature-ir-bricklet|voltage-bricklet|voltage-current-bricklet)
			_tinkerforge_device ${COMP_CWORD} ${prev}
			break
			;;
		*)
			;;
		esac
	done

	return 0
}

complete -F _tinkerforge tinkerforge
